use futures::executor::block_on;
use reqwest::StatusCode;
use std::collections::HashMap;
use std::fs::File;
use std::io::BufRead;

/// Returns a map of the repositories that have a given package as well as the version that they have.
pub fn find_package_in_repositories(package: &str, installed: Option<u64>, ) -> Option<HashMap<String, u64>> {
    let client = reqwest::Client::new(); // The reqwest client which will make GET requests to remove servers.
    let repo_file =
        File::open(crate::config::read_config().installation_path + "repositories.list").expect("Failed to open repository list"); // The repositories.list file, which contains a list of all known and trusted repositories.
    let mut candidates: HashMap<String, u64> = HashMap::new(); // This map will store a list of the repositories along with the latest version of the package they have.

    if installed.is_some() {
        candidates.insert("installed".to_string(), installed.unwrap());
    } // If the package is installed, put the installed version in.
    for repository in std::io::BufReader::new(repo_file).lines() {
        // Iterate through all the repositories in repositories.list
        let repository = repository.unwrap();

        let resp = block_on(
            client
                .get(format!("{repository}/gaypk/{package}/Package.toml.latest"))
                .send(),
        )
        .expect("GET request failed"); // Send a GET request to the current repository, asking for the package's latest known package.toml. TODO protect against version number spoofing

        let status = resp.status(); // Get the status code.
        let text = block_on(resp.text()).expect("Failed to receive packet body"); // Get the message body.

        let package_data = &crate::package::read_package(text); // Parse through TOML to a PackageData.

        if status == StatusCode::from_u16(200).unwrap() {
            // Only add repositories that return 200
            candidates.insert(repository, package_data.revision);
        } else {
            println!("Repository {repository} returned {status}");
        }
    }
    let mut sorted: Vec<_> = candidates.iter().collect(); // Convert candidates into a vector for sorting

    sorted.sort_by_key(|entry| entry.1); // Sort by version number, must be stable to sort in order of priority

    let mut candidates = HashMap::new(); // Clear out the old candidates map
    for (k, v) in sorted.iter() {
        // Iterate through sorted and transfer everything to the new candidates map
        candidates.insert(k.as_str().to_string(), **v); // This is an awful line of code and i hate it
    }

    let mut highest_version: u64 = 0; // The highest version so far
    for (_, version) in candidates.iter() {
        // Iterate through and find the highest version
        if *version > highest_version {
            highest_version = *version
        };
    }

    if candidates.get("installed") == Some(&highest_version) {
        // If the highest version is equal to the installed version,  return None. If not, return the list of candidates for further processing
        None
    } else {
        Some(candidates)
    }
}
